/*
 * ProFTPD: mod_useralias -- looks up aliases for users from berkdb & co.
 *
 * Copyright (c) 2011 Manuel Mausz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 * NOTE:
 * uncomment the line below to use nbdm or gdbm instead of BerkDB
 * you'll need to add another include path via --with-includes
 * also don't forget to modify the $Libraries below
 */
//#define USE_NDBM

/*
 * NOTE: replace "-ldb" with "-lgdbm" for gdbm
 * $Libraries: -ldb $
 */

#define MOD_USERALIAS_VERSION "mod_useralias/0.3.0"

#include "conf.h"
#include "privs.h"
#include <ctype.h>

#if defined(USE_NDBM)
# include <ndbm.h>
#else
# define DB_DBM_HSEARCH 1  /* use the dbm interface */
# define HAVE_DBM          /* for BerkDB 5.0 and later */
# include <db.h>
#endif

static const char *trace_channel = "useralias";
static DBM *dbm = NULL;

module useralias_module;

static int useralias_sess_init(void);

static void db_close()
{
  if (dbm == NULL)
    return;

  pr_trace_msg(trace_channel, 5, "closing database...");
  dbm_close(dbm);
  dbm = NULL;
}

/* authentication handlers */
MODRET useralias_pre_user(cmd_rec *cmd)
{
  datum key, data;
  char *alias, *word, *argdup;
  unsigned char *lower = NULL;
  register unsigned int i;
  array_header *tarr;

  /* check pre-requisites */
  if (dbm == NULL || cmd->argc < 2)
    return PR_DECLINED(cmd);

  alias = cmd->arg;

  pr_trace_msg(trace_channel, 5, "searching for alias '%s' in database", alias);

  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));
  key.dptr  = alias;
  key.dsize = strlen(alias);

  /* convert alias to lower if enabled */
  lower = get_param_ptr(main_server->conf, "UserAliasToLower", FALSE);
  if (lower != NULL && *lower == TRUE)
  {
    for (i = 0; i < key.dsize; ++i)
      key.dptr[i] = tolower(key.dptr[i]);
    pr_trace_msg(trace_channel, 5, "directive 'UserAliasToLower' is enabled. lowered alias is: '%s'", key.dptr);
  }
  else
    pr_trace_msg(trace_channel, 5, "directive 'UserAliasToLower' is disabled");

  /* do the db work */
  pr_trace_msg(trace_channel, 5, "fetching from database: key='%s'...", key.dptr);
  data = dbm_fetch(dbm, key);

  if (data.dptr == NULL)
  {
    pr_log_debug(DEBUG5, MOD_USERALIAS_VERSION ": alias '%s' not found", key.dptr);
    pr_trace_msg(trace_channel, 5, "alias '%s' not found",  key.dptr);
    return PR_DECLINED(cmd);
  }

  pr_log_debug(DEBUG5, MOD_USERALIAS_VERSION ": found alias in database. real user is '%.*s'",
      data.dsize, (char *)data.dptr);
  pr_trace_msg(trace_channel, 5, "found alias in database. real user is '%.*s', len is %d",
      data.dsize, (char *)data.dptr, data.dsize);

  pr_trace_msg(trace_channel, 5, "cmd: arg=%s argv[0]=%s argv[1]=%s argc=%d",
      cmd->arg, (char *)cmd->argv[0], (char *)cmd->argv[1], cmd->argc);

  /* rewrite the values which include our alias */
  cmd->arg = pstrndup(cmd->pool, data.dptr, data.dsize);

  tarr = make_array(cmd->pool, cmd->argc, sizeof(char *));
  *((char **)push_array(tarr)) = cmd->argv[0];
  cmd->argc = 1;

  if (strlen(cmd->arg) > 0)
  {
    argdup = pstrndup(cmd->tmp_pool, data.dptr, data.dsize);
    while ((word = pr_str_get_word(&argdup, PR_STR_FL_PRESERVE_COMMENTS)) != NULL)
    {
      pr_signals_handle();

      *((char **)push_array(tarr)) = pstrdup(cmd->pool, word);
      cmd->argc++;
    }
  }

  *((char **)push_array(tarr)) = NULL;
  cmd->argv = tarr->elts;

  pr_cmd_clear_cache(cmd);

  pr_trace_msg(trace_channel, 5, "cmd: arg=%s argv[0]=%s argv[1]=%s argc=%d",
      cmd->arg, (char *)cmd->argv[0], (char *)cmd->argv[1], cmd->argc);

  return PR_DECLINED(cmd);
}

/* configuration handlers */
MODRET set_useraliastolower(cmd_rec *cmd)
{
  int val = -1;
  config_rec *c = NULL;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);

  val = get_boolean(cmd, 1);
  if (val == -1)
    CONF_ERROR(cmd, "expected Boolean parameter");

  c = add_config_param(cmd->argv[0], 1, NULL);
  c->argv[0] = pcalloc(c->pool, sizeof(unsigned char));
  *((unsigned char *) c->argv[0]) = val;

  return PR_HANDLED(cmd);
}

MODRET set_useraliasdb(cmd_rec *cmd)
{
  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);

  if (pr_fs_valid_path(cmd->argv[1]) < 0)
    CONF_ERROR(cmd, "must be an absolute path");

  add_config_param_str(cmd->argv[0], 1, cmd->argv[1]);
  return PR_HANDLED(cmd);
}

/* event handlers */
static void useralias_shutdown_ev(const void *event_data, void *user_data)
{
  db_close();
}

#if defined(PR_SHARED_MODULE)
static void useralias_mod_unload_ev(const void *event_data, void *user_data)
{
  if (strcmp("mod_useralias.c", (const char *)event_data) == 0)
    pr_event_unregister(&useralias_module, NULL, NULL);
}
#endif

static void useralias_sess_reinit_ev(const void *event_data, void *user_data)
{
  /* A HOST command changed the main_server pointer, reinitialize ourselves. */
  pr_event_unregister(&useralias_module, "core.session-reinit",
      useralias_sess_reinit_ev);

  if (useralias_sess_init() < 0)
    pr_session_disconnect(&useralias_module,
        PR_SESS_DISCONNECT_SESSION_INIT_FAILED, NULL);
}

/* initialization routines */
static int useralias_sess_init(void)
{
  config_rec *c = NULL;
  const char *database = NULL;

  pr_event_register(&useralias_module, "core.session-reinit",
      useralias_sess_reinit_ev, NULL);

  /* search db config value */
  c = find_config(main_server->conf, CONF_PARAM, "UserAliasDB", FALSE);
  if (!c)
  {
    pr_trace_msg(trace_channel, 5, "directive UserAliasDB not set. mod_useralias not enabled");
    return 0;
  }

  database = c->argv[0];

  pr_trace_msg(trace_channel, 5, "trying to open database '%s'", database);
  if (dbm != NULL)
  {
    /* this should never be reached */
    pr_log_pri(PR_LOG_ERR, MOD_USERALIAS_VERSION ": database handle is valid. This should never occur");
    pr_trace_msg(trace_channel, 1, "database handle is valid. This should never occur");
    db_close();
  }

  /* open the database */
  PRIVS_ROOT
  dbm = dbm_open(database, O_RDONLY, 0600);
  PRIVS_RELINQUISH
  if (dbm == NULL)
  {
    pr_log_debug(DEBUG5, MOD_USERALIAS_VERSION ": unable to open database '%s': %m", database);
    pr_trace_msg(trace_channel, 5, "unable to open database '%s': %m", database);
  }

  /* register a shutdown routine to close the db handle */
  pr_event_register(&useralias_module, "core.exit", useralias_shutdown_ev, NULL);
  return 0;
}

static int useralias_init(void)
{
#if defined(PR_SHARED_MODULE)
  pr_event_register(&useralias_module, "core.module-unload",
      useralias_mod_unload_ev, NULL);
#endif
  return 0;
}

/* module api tables */
static conftable useralias_conftab[] =
{
  { "UserAliasDB",      set_useraliasdb,      NULL },
  { "UserAliasToLower", set_useraliastolower, NULL },
  { NULL }
};

static cmdtable useralias_cmdtab[] =
{
  { PRE_CMD, C_USER, G_NONE, useralias_pre_user, FALSE,  FALSE, CL_AUTH },
  { 0, NULL }
};

module useralias_module =
{
  /* always NULL */
  NULL, NULL,

  /* module api version 2.0 */
  0x20,

  /* module name */
  "useralias",

  /* module configuration handler table */
  useralias_conftab,

  /* module command handler table */
  useralias_cmdtab,

  /* module authentication handler table */
  NULL,

  /* module initialization function */
  useralias_init,

  /* module session initialization function */
  useralias_sess_init,

  /* module version */
  MOD_USERALIAS_VERSION
};
